/* test.c
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023, 2024 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#include <assert.h>
#include <stdarg.h>
#include <stdbool.h>

#include "test.h"

#include <pthread.h>
void thread_construct(thread *this)
{
	this->tid = 0u;
}

void thread_destruct(thread *this)
{
}

pthread_mutex_t * new_pthread_mutex_t_must()
{
	pthread_mutex_t *mutexnbl = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
	if(mutexnbl) {
		pthread_mutex_t *mutexnn = (pthread_mutex_t *) mutexnbl;

		return mutexnn;
	}

	perror(NULL);
	exit(EXIT_FAILURE);
}

void thread_mutex_construct(thread_mutex *this)
{
	this->mutex = new_pthread_mutex_t_must();
	pthread_mutex_init(this->mutex, NULL);
}

void thread_mutex_destruct(thread_mutex *this)
{
}

thread * thread_start(void * (*twfun)(void *usrdata), void *usrdata)
{
	unsigned long tid = 0u;
	int retval = 0;
	retval = pthread_create(&tid, NULL, twfun, usrdata);
	if(retval != 0) {
		return NULL;
	}

	thread *th = (thread *) malloc(sizeof(thread));
	if(th == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	thread_construct(th);
	th->tid = tid;

	return th;
}

thread * thread_start_or_serial(void * (*twfun)(void *usrdata), void *usrdata)
{
	thread *th = NULL;
	th = thread_start(twfun, usrdata);
	if(NULL == th) {
		twfun(usrdata);
	}


	return th;
}

void thread_join(thread *thandle)
{
	if(NULL != thandle) {
		pthread_join(thandle->tid, NULL);
	}

}

void thread_mutex_lock(thread_mutex *thandle)
{
	pthread_mutex_lock(thandle->mutex);
}

void thread_mutex_trylock(thread_mutex *thandle)
{
	pthread_mutex_trylock(thandle->mutex);
}

void thread_mutex_unlock(thread_mutex *thandle)
{
	pthread_mutex_unlock(thandle->mutex);
}

int thread_cancel(thread *thandle)
{
	return pthread_cancel(thandle->tid);
}

void uidelay()
{
	sleep(2u);
}

void uidelay_minimal()
{
	usleep((33u) * (1000u));
}

int helper_select_next_prebrush(myapp *app, int i)
{
	assert(i >= 0); /* test.ngg:34 */
	switch(i) {
	case 0:
	{
		g_main_context_invoke(NULL, (void *) mctest_select_ballpoint_brush, app);
		break;
	}
	case 1:
	{
		g_main_context_invoke(NULL, (void *) mctest_select_charcoal_brush, app);
		break;
	}
	case 2:
	{
		g_main_context_invoke(NULL, (void *) mctest_select_round_brush, app);
		break;
	}
	case 3:
	{
		g_main_context_invoke(NULL, (void *) mctest_select_round_pressure_size_brush, app);
		break;
	}
	case 4:
	{
		g_main_context_invoke(NULL, (void *) mctest_select_soft_round_brush, app);

		return -1;
		break;
	}
	}


	return i + 1;
}

_Bool mctest_increase_fill_size(myapp *app)
{
	gtk_widget_activate((GtkWidget *) app->mni_tools_bplus);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_ballpoint_brush(myapp *app)
{
	gtk_button_clicked(app->btn_prebrush_ballpoint);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_brush_tool(myapp *app)
{
	gtk_toggle_button_set_active((GtkToggleButton *) app->btn_brush, true);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_charcoal_brush(myapp *app)
{
	gtk_button_clicked(app->btn_prebrush_charcoal);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_floodfill_tool(myapp *app)
{
	gtk_toggle_button_set_active((GtkToggleButton *) app->btn_floodfill, true);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_round_brush(myapp *app)
{
	gtk_button_clicked(app->btn_prebrush_round);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_randsquare_brush_type(myapp *app)
{
	gtk_combo_box_set_active_id((GtkComboBox *) app->cbo_brush_engine, "Rand. Square");

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_round_pressure_size_brush(myapp *app)
{
	gtk_button_clicked(app->btn_prebrush_round_pressure_size);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_select_soft_round_brush(myapp *app)
{
	gtk_button_clicked(app->btn_prebrush_soft_round);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_assert_round_brush(myapp *app)
{
	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_zoom_in(myapp *app)
{
	if(gtk_widget_get_sensitive((GtkWidget *) app->mni_view_zoomin)) {
		assert(gtk_widget_activate((GtkWidget *) app->mni_view_zoomin)); /* test.ngg:118 */
	}


	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_zoom_out(myapp *app)
{
	if(gtk_widget_get_sensitive((GtkWidget *) app->mni_view_zoomout)) {
		assert(gtk_widget_activate((GtkWidget *) app->mni_view_zoomout)); /* test.ngg:128 */
	}


	return (_Bool) G_SOURCE_REMOVE;
}

void test_stroke_context_construct(TestStrokeContext *this)
{
	this->xoffset = 0;
	this->maxstp = 0;
	this->stp = 0;
	this->app = NULL;
}

void test_stroke_context_destruct(TestStrokeContext *this)
{
}

_Bool mctest_draw(TestStrokeContext *ctx)
{
	assert(ctx->stp >= 0); /* test.ngg:141 */
	assert(ctx->stp <= ctx->maxstp); /* test.ngg:142 */
	double xoffset = ctx->app->doc->vpx + ctx->xoffset;
	int yoffset = ctx->app->doc->vpy;
	int x = (int) (xoffset + ((int) (sin(6.28 * (((double) ctx->stp) / ctx->maxstp)) * 20)));
	int y = (int) (yoffset + ((int) (ctx->stp * 20)));
	int halfstps = ctx->maxstp / 2;
	double pressure = 1 - (fabs(halfstps - ctx->stp) / halfstps);
	int event_type = GDK_MOTION_NOTIFY;
	if(ctx->stp == 0) {
		event_type = GDK_BUTTON_PRESS;
	} else if(ctx->stp == ctx->maxstp) {
		event_type = GDK_BUTTON_RELEASE;
	}

	myapp_process_da_event(ctx->app, event_type, x, y, pressure, GDK_SCROLL_UP);
	myapp_drawframe_noskip(ctx->app);
	ctx->stp += 1;

	return (_Bool) G_SOURCE_REMOVE;
}

void test_draw(myapp *app, int xoffset, _Bool animate)
{
	int i;
	TestStrokeContext *ctx = (TestStrokeContext *) malloc(sizeof(TestStrokeContext));
	if(ctx == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	test_stroke_context_construct(ctx);
	ctx->app = app;
	ctx->maxstp = 20;
	ctx->xoffset = xoffset;
	for(i = 0; i <= ctx->maxstp; i += 1) {
		g_main_context_invoke(NULL, (void *) mctest_draw, ctx);
		if(animate) {
			uidelay_minimal();
		}

	}

	if(ctx) {
		test_stroke_context_destruct(ctx);
		free(ctx);
	}

}

void mchelper_draw_dot_perc(myapp *app, double x, double y, double pressure)
{
	mchelper_draw_line_perc(app, x, y, pressure, x, y, pressure);
}

void mchelper_draw_line_perc(myapp *app, double x1, double y1, double p1, double x2, double y2, double p2)
{
	app->doc->drawparams.oldpressure = p1;
	app->doc->drawparams.pressure = p2;
	struct Rectangle rect = image_line(app->doc->layer->img, x1 * image_document_get_width(app->doc), y1 * image_document_get_height(app->doc), x2 * image_document_get_width(app->doc), y2 * image_document_get_height(app->doc), app->doc->drawparams);
	image_document_add_dirty_region(app->doc, rect);
}

void mchelper_set_fill_size_perc(myapp *app, double siz)
{
	app->doc->drawparams.brushparams.siz = (int) (siz * image_document_get_width(app->doc));
}

void mchelper_set_stroke_color(myapp *app, double r, double g, double b)
{
	app->doc->drawparams.fgcolor = nan_pixel_new(r, g, b, 1);
}

void mchelper_draw_vgradient(myapp *app, double yfrom, double yto, double stp, double rfac, double gfac, double bfac, double opacity, _Bool absolutergb)
{
	int y;
	assert(yfrom <= yto); /* test.ngg:228 */
	int iyfrom = (int) (yfrom * 100);
	int iyto = (int) (yto * 100);
	int istp = (int) (stp * 100);
	mctest_select_soft_round_brush(app);
	app->doc->drawparams.brushparams.opacity *= opacity;
	mchelper_set_fill_size_perc(app, 1.5 * stp);
	for(y = iyfrom; y <= iyto; y += istp) {
		double loopfac = (y - iyfrom) / (((double) iyto) - iyfrom);
		double yfac = y / 100.0;
		printf("yfac: %f, loopfac: %f\n", yfac, loopfac);
		if(absolutergb) {
			mchelper_set_stroke_color(app, rfac, gfac, bfac);
		} else {
			mchelper_set_stroke_color(app, rfac * loopfac, gfac * loopfac, bfac * loopfac);
		}

		mchelper_draw_line_perc(app, 0, yfac, 1, 1, yfac, 1);
		myapp_drawframe(app);
	}

}

_Bool mctest_draw_icons(myapp *app)
{
	int iconsiz = 28;
	double padding = 0.05;
	double stroke_width = 0.1;
	double fill_size = 0.3;
	ImageDocument *_tmp_2 = (ImageDocument *) malloc(sizeof(ImageDocument));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_document_construct(_tmp_2, iconsiz, iconsiz);
	myapp_set_document(app, _tmp_2);
	gtk_toggle_button_set_active((GtkToggleButton *) app->btn_brush, true);
	mctest_select_round_brush(app);
	app->doc->drawparams.brushparams.pressure_size = true;
	double pencilnecky = 0.8;
	double pencil_fill_size = fill_size * 0.8;
	image_document_clear_layer(app->doc);
	mchelper_set_fill_size_perc(app, pencil_fill_size + stroke_width);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_draw_line_perc(app, 0.5, pencilnecky, 1, 0.5, padding, 0);
	mchelper_draw_line_perc(app, 0.5, pencilnecky, 1, 0.5, 1, 1);
	mchelper_set_fill_size_perc(app, pencil_fill_size);
	mchelper_set_stroke_color(app, 0.5, 0.5, 0.5);
	mchelper_draw_line_perc(app, 0.5, pencilnecky, 1, 0.5, 1, 1);
	mchelper_set_stroke_color(app, 0.7, 0.7, 0.7);
	mchelper_draw_line_perc(app, 0.5, pencilnecky, 1, 0.5, padding + stroke_width, 0);
	mchelper_set_fill_size_perc(app, 0.15);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_draw_line_perc(app, 0.5, padding + stroke_width + 0.3, 1, 0.5, padding + stroke_width, 0);
	mctest_set_button_icon_from_fg(app, (GtkButton *) app->btn_pencil);
	double brush_gap = stroke_width * 1.2;
	double handle_size = fill_size * 0.5;
	image_document_clear_layer(app->doc);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_set_fill_size_perc(app, handle_size + stroke_width);
	mchelper_draw_line_perc(app, 0.5, pencilnecky + brush_gap, 1, 0.5, 1, 1);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_set_fill_size_perc(app, fill_size + stroke_width);
	mchelper_draw_line_perc(app, 0.5, pencilnecky - brush_gap, 1, 0.5, padding, 0);
	mchelper_set_stroke_color(app, 0.5, 0.5, 0.5);
	mchelper_set_fill_size_perc(app, handle_size);
	mchelper_draw_line_perc(app, 0.5, pencilnecky + brush_gap, 1, 0.5, 1, 1);
	mchelper_set_stroke_color(app, 0, 0.5, 1);
	mchelper_set_fill_size_perc(app, fill_size);
	mchelper_draw_line_perc(app, 0.5, pencilnecky - brush_gap, 1, 0.5, padding + stroke_width, 0);
	mctest_set_button_icon_from_fg(app, (GtkButton *) app->btn_brush);
	double eraser_fill_size = fill_size * 1.5;
	double erasertopy = (fill_size / 2.0) + padding + stroke_width;
	image_document_clear_layer(app->doc);
	mchelper_set_fill_size_perc(app, eraser_fill_size);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_draw_line_perc(app, 0.5, erasertopy, 1, 0.5, 1 - erasertopy, 1);
	mchelper_set_fill_size_perc(app, eraser_fill_size - stroke_width);
	mchelper_set_stroke_color(app, 0.5, 0.5, 0.5);
	mchelper_draw_line_perc(app, 0.5, erasertopy, 1, 0.5, 0.5, 1);
	mchelper_set_stroke_color(app, 1, 1, 1);
	mchelper_draw_line_perc(app, 0.5, 0.5, 1, 0.5, 1 - erasertopy, 1);
	mctest_set_button_icon_from_fg(app, (GtkButton *) app->btn_eraser);
	image_document_clear_layer(app->doc);
	double eyedropper_fill_size = pencil_fill_size;
	double eyedroppernecky = pencilnecky - 0.5;
	double eyedroppertipy = pencilnecky - 0.2;
	double eyedropperdroptop = eyedroppertipy + 0.01;
	double eyedropper_fill_radius = eyedropper_fill_size * 0.5;
	double eyedropperdropend = 1 - eyedropper_fill_radius - stroke_width;
	mchelper_set_fill_size_perc(app, eyedropper_fill_size + stroke_width);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_draw_line_perc(app, 0.5, eyedroppernecky, 1, 0.5, 0, 1);
	mchelper_draw_line_perc(app, 0.5, eyedroppernecky, 1, 0.5, eyedroppertipy + stroke_width, 0);
	mchelper_draw_line_perc(app, 0.5, eyedropperdroptop, 0, 0.5, eyedropperdropend, 1);
	mchelper_set_fill_size_perc(app, pencil_fill_size);
	mchelper_set_stroke_color(app, 0.5, 0.5, 0.5);
	mchelper_draw_line_perc(app, 0.5, eyedroppernecky, 1, 0.5, 0, 1);
	mchelper_draw_line_perc(app, 0.5, eyedroppernecky, 1, 0.5, eyedroppertipy, 0);
	mchelper_set_stroke_color(app, 0, 0.5, 1);
	mchelper_draw_line_perc(app, 0.5, eyedropperdroptop, 0, 0.5, eyedropperdropend, 1);
	mctest_set_button_icon_from_fg(app, (GtkButton *) app->btn_eyedropper);
	image_document_clear_layer(app->doc);
	double bucket_fill_size = 0.6;
	mchelper_set_fill_size_perc(app, bucket_fill_size);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_draw_line_perc(app, 0.4, erasertopy, 1, 0.6, 1 - erasertopy, 1);
	mchelper_set_fill_size_perc(app, bucket_fill_size - stroke_width);
	mchelper_set_stroke_color(app, 0.5, 0.5, 0.5);
	mchelper_draw_line_perc(app, 0.4, erasertopy, 1, 0.6, 1 - erasertopy, 1);
	mchelper_set_fill_size_perc(app, bucket_fill_size);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_draw_line_perc(app, 0.4, erasertopy, 1, 0.4, erasertopy, 1);
	mchelper_set_fill_size_perc(app, bucket_fill_size - stroke_width);
	mchelper_set_stroke_color(app, 0, 0.5, 1);
	mchelper_draw_line_perc(app, 0.4, erasertopy, 1, 0.4, erasertopy, 1);
	gtk_toggle_button_set_active((GtkToggleButton *) app->btn_brush, true);
	mchelper_set_stroke_color(app, 0, 0, 0);
	mchelper_set_fill_size_perc(app, eyedropper_fill_size + stroke_width);
	mchelper_draw_line_perc(app, 0.3, 0.48, 0, 0.3, 0.8, 1);
	mchelper_set_stroke_color(app, 0, 0.5, 1);
	mchelper_set_fill_size_perc(app, pencil_fill_size);
	mchelper_draw_line_perc(app, 0.3, 0.45, 0, 0.3, 0.8, 1);
	mctest_set_button_icon_from_fg(app, (GtkButton *) app->btn_floodfill);
	myapp_reflect_brush_status(app);
	myapp_drawframe_noskip(app);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_draw_sunset(myapp *app)
{
	int i;
	double suny = 0.41;
	double sunsiz = 0.1;
	gtk_toggle_button_set_active((GtkToggleButton *) app->btn_brush, true);
	mchelper_draw_vgradient(app, 0, 1, 0.1, 1, 0.2, 0, 1, false);
	mctest_select_soft_round_brush(app);
	app->doc->drawparams.brushparams.pressure_size = true;
	mchelper_set_fill_size_perc(app, 0.2);
	mchelper_set_stroke_color(app, 0.3, 0, 0);
	mchelper_draw_line_perc(app, 0, suny, 1, 0.45, suny, 0);
	mchelper_draw_line_perc(app, 0.55, suny, 0, 1, suny, 1);
	mchelper_set_fill_size_perc(app, 0.1);
	mchelper_set_stroke_color(app, 0.1, 0, 0);
	mchelper_draw_line_perc(app, 0, suny, 1, 0.45, suny, 0);
	mchelper_draw_line_perc(app, 0.55, suny, 0, 1, suny, 1);
	mchelper_set_fill_size_perc(app, 0.5);
	mchelper_set_stroke_color(app, 1, 1, 0);
	mchelper_draw_dot_perc(app, 0.5, 0.4, 1);
	mchelper_set_fill_size_perc(app, 0.3);
	mchelper_set_stroke_color(app, 1, 1, 0);
	mchelper_draw_dot_perc(app, 0.5, 0.4, 1);
	mctest_select_round_brush(app);
	mchelper_set_fill_size_perc(app, sunsiz);
	mchelper_set_stroke_color(app, 1, 0.8, 0);
	mchelper_draw_dot_perc(app, 0.5, 0.4, 1);
	app->doc->drawparams.brushparams.opacity = 1;
	mchelper_draw_vgradient(app, 0.45, 1, 0.05, 0.5, 0.1, 0, 1, false);
	mchelper_draw_vgradient(app, 0.45, 1, 0.05, 0.4, 0.2, 0, 0.1, true);
	mctest_select_charcoal_brush(app);
	double sunrad = sunsiz / 2.0;
	for(i = 0; i <= 60; i += 2) {
		double yfac = i / 60.0;
		double finaly = suny + (yfac * (yfac * 0.6));
		double xshift = (finaly - suny) / 20.0;
		printf("yfac: %f\n", yfac);
		mchelper_set_stroke_color(app, 1, 0.8, 0);
		mchelper_set_fill_size_perc(app, 0.01);
		mchelper_draw_line_perc(app, 0.5 - sunrad - xshift, finaly, 1, 0.5 + sunrad + xshift, finaly, 1);
		mchelper_set_stroke_color(app, 0, 0, 0);
		double dxstp = 0.01 + (yfac / 10);
		int ixstp = (int) (dxstp * 100);
		if(ixstp > 0) {
			int x;
			int alter = 1;
			for(x = 0; x < 100; x += ixstp) {
				double curx = (x / 100.0) + (alter * dxstp);
				double cury = finaly + (yfac * (alter * 0.001));
				mchelper_draw_line_perc(app, curx, cury, 1, curx + dxstp, cury, 0.5);
				alter *= -1;
			}

		}

	}

	myapp_reflect_brush_status(app);
	myapp_drawframe_noskip(app);

	return (_Bool) G_SOURCE_REMOVE;
}

_Bool mctest_draw_vara_logo(myapp *app)
{
	gtk_toggle_button_set_active((GtkToggleButton *) app->btn_brush, true);
		int iconsiz = app->test_vara_logo_size;
		double padding = 0.05;
		double stroke_width = 0.1;
		double fill_size = 0.3;
		ImageDocument *_tmp_3 = (ImageDocument *) malloc(sizeof(ImageDocument));
		if(_tmp_3 == NULL) {
			perror(NULL);
			exit(EXIT_FAILURE);
		}
	
		image_document_construct(_tmp_3, iconsiz, iconsiz);
		myapp_set_document(app, _tmp_3);
	
		double pencilnecky = 0.8;
		double brush_gap = stroke_width * 1.2;
		double handle_size = fill_size * 0.5;
		image_document_clear_layer(app->doc);
		
		// Margin - from image bound
		double bezel_margin_x = 0.1;
		double bezel_margin_y = 0.2;
		double bezel_offset_y = -0.1;
		double bezel_height = 0.6;
	
		// The reason x doesn't get this complicated is that width of the bezel
		// is unspecified.
		double bezel_y_bottom = bezel_margin_y + bezel_height + bezel_offset_y;
		
		image_fill_region(app->doc->fglayer->img, PIXEL_WHITE,
			rectangle_new_from_boundaries(bezel_margin_x * iconsiz, (bezel_margin_y + bezel_offset_y) * iconsiz, iconsiz - (bezel_margin_x * iconsiz), bezel_y_bottom * iconsiz));
		
		mctest_select_round_brush(app);
		mchelper_set_fill_size_perc(app, stroke_width);
		app->doc->drawparams.brushparams.pressure_size = true;
	
		double drop_shadow_offset_y = 0.02;
	
		// Bottom shadow
		mchelper_set_stroke_color(app, 0.4, 0.4, 0.4);
		mchelper_draw_line_perc(app, bezel_margin_x, bezel_y_bottom + drop_shadow_offset_y, 1, 1 - bezel_margin_x, bezel_y_bottom + drop_shadow_offset_y, 1);
	
		mchelper_set_stroke_color(app, 0.6, 0.6, 0.6);
	
		// Top
		mchelper_draw_line_perc(app, bezel_margin_x, bezel_margin_y + bezel_offset_y, 1, 1 - bezel_margin_x, bezel_margin_y + bezel_offset_y, 1);
		// Bottom
		mchelper_draw_line_perc(app, bezel_margin_x, bezel_y_bottom, 1, 1 - bezel_margin_x, bezel_y_bottom, 1);
		// Left
		mchelper_draw_line_perc(app, bezel_margin_x, bezel_margin_y + bezel_offset_y, 1, bezel_margin_x, bezel_y_bottom, 1);
		// Right
		mchelper_draw_line_perc(app, 1 - bezel_margin_x, bezel_margin_y + bezel_offset_y, 1, 1 - bezel_margin_x, bezel_y_bottom, 1);
	
		// Palette
	
	/*	double colors[][3] = {*/
	/*		{ 1, 0, 0 },*/
	/*		{ 1, 1, 0 },*/
	/*		{ 0, 1, 0 },*/
	/*		{ 0, 1, 1 },*/
	/*		{ 0, 0, 1 },*/
	/*	};*/
		
		double colors[][3] = {
			{ 1, 0, 0 },
			{ 1, 1, 0 },
			{ 0, 1, 0 },
			{ 0, 0.6, 1 },
			{ 1, 0, 1 },
		};
		
		const int ncolors = sizeof(colors) / sizeof(colors[1]);
	
		double palette_dot_size = 1.0 / (ncolors + 1);
		mchelper_set_fill_size_perc(app, palette_dot_size);
	
		double palette_startx = 0.5 - ((palette_dot_size * ncolors) / 2) + (palette_dot_size / 2);
		for(int i = 0; i < ncolors; i++) {
			double x = palette_startx + i * palette_dot_size;
			
			mchelper_set_stroke_color(app, 0.4, 0.4, 0.4);
			mchelper_draw_line_perc(app, x, bezel_y_bottom + palette_dot_size + drop_shadow_offset_y, 1, x, bezel_y_bottom + palette_dot_size + drop_shadow_offset_y, 1);
	
			mchelper_set_stroke_color(app, colors[i][0], colors[i][1], colors[i][2]);
			mchelper_draw_line_perc(app, x, bezel_y_bottom + palette_dot_size, 1, x, bezel_y_bottom + palette_dot_size, 1);
		}
	
		// Slanted Brush
	
		pencilnecky = bezel_y_bottom - 0.1;
		stroke_width = stroke_width * 0.5;
		
		mchelper_set_stroke_color(app, 0.25, 0.25, 0.25);
		mchelper_set_fill_size_perc(app, handle_size + stroke_width);
		mchelper_draw_line_perc(app, 0.2, 0.6, 1, -0.4, 1, 0.6);
	
		// TODO REM one
		mchelper_set_stroke_color(app, 0.3, 0.2, 0.1); // brown
		mchelper_set_stroke_color(app, 0, 0.2, 0.6);
		mchelper_set_fill_size_perc(app, fill_size + stroke_width);
		//mchelper_draw_line_perc(app, 0.4, pencilnecky - brush_gap, 1, 0.8, padding + 0.1, 0);
		// With bent
		double endx = 0.7;
		double endy = padding + 0.2;
		double endpr = 0.2;
		mchelper_draw_line_perc(app, 0.4, pencilnecky - brush_gap, 1, endx, endy, endpr);
		mchelper_draw_line_perc(app, endx, endy, endpr, endx + 0.04, endy - 0.06, 0);
		
		mchelper_set_stroke_color(app, 0.5, 0.5, 0.5);
		mchelper_set_fill_size_perc(app, handle_size);
		mchelper_draw_line_perc(app, 0.2, 0.6, 1, -0.4, 1, 0.4);
	
		mctest_select_soft_round_brush(app);
		mchelper_set_stroke_color(app, 0.8, 0.8, 0.8);
	
		// Thin highlight for 3D look
		//mchelper_set_fill_size_perc(app, handle_size * 0.5);
		//mchelper_draw_line_perc(app, 0.2, 0.55, 1, -0.4, 0.95, 0.3);
	
		// GNOME-style flat 3D look
		mchelper_set_fill_size_perc(app, handle_size);
		mchelper_draw_line_perc(app, 0.2, 0.59, 1, -0.4, 0.99, 0.3);
	
		// Something in-between
		//mchelper_set_fill_size_perc(app, handle_size * 0.9);
		//mchelper_draw_line_perc(app, 0.2, 0.58, 1, -0.4, 0.98, 0.3);
	
		mctest_select_round_brush(app);
		app->doc->drawparams.brushparams.pressure_size = true;
		// TODO REM one
		mchelper_set_stroke_color(app, 0.7, 0.5, 0.2); // brown
		mchelper_set_stroke_color(app, 0, 0.5, 1);
		mchelper_set_fill_size_perc(app, fill_size);
		//mchelper_draw_line_perc(app, 0.4, pencilnecky - brush_gap, 1, 0.8, padding + 0.1, 0);
		// With bent
		mchelper_draw_line_perc(app, 0.4, pencilnecky - brush_gap, 1, endx, endy, endpr);
		mchelper_draw_line_perc(app, endx, endy, endpr, endx + 0.04, endy - 0.06, 0);
	
		mctest_select_soft_round_brush(app);
		mchelper_set_stroke_color(app, 0.8, 0.8, 0.8);
		mchelper_set_fill_size_perc(app, fill_size);
		mchelper_draw_line_perc(app, 0.4, pencilnecky - brush_gap, 1, 0.4 - 0.04, pencilnecky - brush_gap - 0.025, 0);
	
		// Clear the left padding (there is the handle coming out)
		app->doc->drawparams.brushparams.pressure_size = false;
		app->doc->drawparams.brushparams.opacity = 1;
		app->doc->drawparams.brushparams.softness = 0;
		app->doc->drawparams.erase = true;
		mchelper_set_fill_size_perc(app, padding);
		mchelper_draw_line_perc(app, 0, 0, 1, 0, 1, 1);
	
		image_document_export_layer(app->doc, "/tmp/vara-logo.png", false);
	
		if(app->test_vara_logo_exit) {
			exit(EXIT_SUCCESS);
		}
	
		myapp_reflect_brush_status(app);
		myapp_drawframe_noskip(app);
	

	return (_Bool) G_SOURCE_REMOVE;
}

void mctest_set_button_icon_from_fg(myapp *app, GtkButton *btn)
{
	ImageView *iv = (ImageView *) malloc(sizeof(ImageView));
	if(iv == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_view_construct(iv);
	image_view_set_image(iv, image_make_clone(app->doc->fglayer->img));
	GtkEventBox *ebx = (GtkEventBox *) gtk_event_box_new();
	gtk_widget_add_events((GtkWidget *) ebx, GDK_POINTER_MOTION_MASK);
	GtkWidget *_tmp_2 = image_view_get_native_widget(iv);
	gtk_container_add(GTK_CONTAINER((GtkWidget *) ebx), GTK_WIDGET(_tmp_2));
	gtk_button_set_image(btn, (GtkWidget *) ebx);
	gtk_button_set_label(btn, NULL);
	gtk_button_set_relief(btn, GTK_RELIEF_NONE);
}

void test_brushes(myapp *app)
{
	g_main_context_invoke(NULL, (void *) mctest_select_brush_tool, app);
	int xoffset = 100;
	int preset = 0;
	do {
		int x1i;
		preset = helper_select_next_prebrush(app, preset);
		uidelay();
		xoffset = test_next_xoffset(xoffset, app);
		test_draw(app, xoffset, true);
		double coords[3] = {-0.1, 0.5, 1.1};
		for(x1i = 0; x1i < (sizeof coords / sizeof coords[0]); x1i += 1) {
			int y1i;
			for(y1i = 0; y1i < (sizeof coords / sizeof coords[0]); y1i += 1) {
				int x2i;
				for(x2i = 0; x2i < (sizeof coords / sizeof coords[0]); x2i += 1) {
					int y2i;
					for(y2i = 0; y2i < (sizeof coords / sizeof coords[0]); y2i += 1) {
						mchelper_draw_line_perc(app, coords[x1i], coords[y1i], 1, coords[x2i], coords[y2i], 1);
					}

				}

			}

		}

	} while(preset > 0);

}

int test_next_xoffset(int xoffset, myapp *app)
{
	xoffset += 50;
	if(xoffset > app->doc->fglayer->img->width) {
		xoffset = 20;
	}


	return xoffset;
}

void test_nocrash_brushes_while_floodfill(myapp *app)
{
	g_main_context_invoke(NULL, (void *) mctest_select_floodfill_tool, app);
	uidelay();
	int preset = 0;
	do {
		preset = helper_select_next_prebrush(app, preset);
		uidelay();
	} while(preset > 0);

}

void testmain_general(myapp *app)
{
	thread *thrd = thread_start((void * (*)(void *usrdata)) tw_testmain_general, app);
	assert(thrd); /* test.ngg:798 */
}

void testmain_draw_icons(myapp *app)
{
	thread *thrd = thread_start((void * (*)(void *usrdata)) tw_testmain_draw_icons, app);
	assert(thrd); /* test.ngg:803 */
}

void testmain_draw_sunset(myapp *app)
{
	thread *thrd = thread_start((void * (*)(void *usrdata)) tw_testmain_draw_sunset, app);
	assert(thrd); /* test.ngg:808 */
}

void testmain_draw_vara_logo(myapp *app)
{
	thread *thrd = thread_start((void * (*)(void *usrdata)) tw_testmain_draw_vara_logo, app);
	assert(thrd); /* test.ngg:813 */
}

void tw_testmain_general(myapp *app)
{
	int j;
	int i;
	uidelay();
	test_nocrash_brushes_while_floodfill(app);
	test_brushes(app);
	for(i = 0; i < 10; i += 1) {
		g_main_context_invoke(NULL, (void *) mctest_increase_fill_size, app);
		uidelay();
	}

	for(i = 0; i < 20; i += 1) {
		g_main_context_invoke(NULL, (void *) mctest_zoom_in, app);
		uidelay();
	}

	for(i = 0; i < 40; i += 1) {
		g_main_context_invoke(NULL, (void *) mctest_zoom_out, app);
		uidelay();
	}

	for(j = 0; j < 100; j += 1) {
		printf("draw %d\n", j);
		test_draw(app, 100, true);
	}

	puts("ok.");
	exit(EXIT_SUCCESS);
}

void tw_testmain_draw_icons(myapp *app)
{
	uidelay();
	g_main_context_invoke(NULL, (void *) mctest_draw_icons, app);
}

void tw_testmain_draw_sunset(myapp *app)
{
	uidelay();
	g_main_context_invoke(NULL, (void *) mctest_draw_sunset, app);
}

void tw_testmain_draw_vara_logo(myapp *app)
{
	uidelay_minimal();
	g_main_context_invoke(NULL, (void *) mctest_draw_vara_logo, app);
}

